import processing.core.*; 

import controlP5.*; 

import java.applet.*; 
import java.awt.Dimension; 
import java.awt.Frame; 
import java.awt.event.MouseEvent; 
import java.awt.event.KeyEvent; 
import java.awt.event.FocusEvent; 
import java.awt.Image; 
import java.io.*; 
import java.net.*; 
import java.text.*; 
import java.util.*; 
import java.util.zip.*; 
import java.util.regex.*; 

public class BirdFlocking extends PApplet {

//INTERACTIVE VISUALIZATION OF BIRD FLOCKING, Markus Selmke (selmke@rz.uni-leipzig.de, markus.selmke@gmx.de)


ControlP5 controlP5;

//image variables (i.e. .png formulas, particle images and background)
PImage imCH, imphip, imphil, imphin, imv0, imBIRD, imR, imdv;

//physics parameters
public float phip=0.12f, phil=0.5f, phin, R=5, pR=50, Vel=3, WallPotWidth=50, ArrLength=20, vbase=7, alphainit=0; 
public float Theta, alph, dTheta, dThetah, thetakp, thetakm, xp=0, yp=0, rp=0, thetap=0, thetam=0, bdifx, bdify, xc, yc, thetak, randangle; 
public float projangle=0, projanglex, projangley, opacity, vlocx, vlocy, vlocangle, veloangle, xcentroid, ycentroid, vkActx, vkActy, vkAct, rcMax, rc, imBIRDWidthH, imBIRDHeightH;
public int XSize=1000, YSize=740, bw=100, bh=16, kAct, kover, boverFC, DoubleClick;
int i=0, j=0, k=0, SC, kstart, kend, iNNc, CylceC;
public PFont font, CP5font;

//toggles for simulation and live-plotting, switched by user interface
public boolean bover, locked, visible, iNN;
public boolean AnnotToggle=true, ProjToggle=false, PredatorToggle, ArrowToggle=true, RUN=false, ShadowsToggle=true, DrawBird=false, SpheresToggle=true, DrawCentroid=true, FastToggle=false;

public int Res=360, NP=50, nNN=4;
public int[][] Shadow = new int[50][360]; //[Nparticles][Res], array stores Shadows
public float[][] Particles = new float[50][4]; //[Nparticles][4], {x,y,vx,vy}
public float[] rpVec = new float[50]; //[Nparticles], {rp}
public float[] rpSort = new float[50]; //[Nparticles], {rp}
public int[] iorder = new int[50]; //[Nparticles]
//PImage Density = createImage(1000,700,ALPHA);  // Make a PImage object

public void setup() { 
  size(1000, 740);
  //smooth();
  //frameRate(400);
  
  dTheta = 2*PI/Res;
  dThetah = dTheta/2;
  CylceC = 0;
  xcentroid = XSize/2;
  ycentroid = YSize/2;
  phin = 1 - phil - phip;
  
  //initial particle position
  ClearShadow();
  InitParticles();
  
  // The background image must be the same size as the parameters into the size() method.
  // is 800 x 800 pixels. Load all used images.
  imCH = loadImage("crosshaircyan.png");
  imphip = loadImage("im_phip.png");
  imphil = loadImage("im_phil.png");
  imphin = loadImage("im_phin.png");
  imv0 = loadImage("im_v0.png");
  imBIRD = loadImage("bee3s.png");
  //imBIRD = loadImage("bird3.png");
  imR = loadImage("im_R.png");
  imdv = loadImage("im_dv.png");
  
  imBIRDWidthH = imBIRD.width / 2;
  imBIRDHeightH = imBIRD.height / 2;
  
  font = createFont("Arial", 12); 
  textFont(font);
  
  //initialize control sliders and checkboxes
    controlP5 = new ControlP5(this);
    ControlFont CP5font = new ControlFont(font); 
    controlP5.setControlFont(CP5font);
  //Sliders for changable parameters
    Slider Veloslider = controlP5.addSlider("velocity", 0, 20, 0, 10, 10, bw, bh);
      Veloslider.setValue(Vel);
    Slider Phipslider = controlP5.addSlider("phip", 0, 1, 0, 10, 30, bw, bh);
      Phipslider.setValue(phip);
    Slider Philslider = controlP5.addSlider("phil", 0, 1, 0, 10, 50, bw, bh);
      Philslider.setValue(phil);
    Slider Rslider = controlP5.addSlider("Rad", 0.1f, 20, 0, 10, 90, bw, bh);
      Rslider.setValue(R);
    Slider vbaseslider = controlP5.addSlider("vbase", 0.1f, 20, 0, 10, 110, bw, bh);
      vbaseslider.setValue(vbase);
      
  //change their apperance!
    Veloslider.captionLabel().toUpperCase(false);
    Phipslider.captionLabel().toUpperCase(false);
    Philslider.captionLabel().toUpperCase(false);
    Veloslider.setCaptionLabel(" ");
    Phipslider.setCaptionLabel(" ");
    Philslider.setCaptionLabel(" ");
    Rslider.setCaptionLabel(" ");
    vbaseslider.setCaptionLabel(" ");
      
}

public void velocity(float theValue) {
  Vel=theValue;
}

public void phip(float theValue) {
  phip=theValue;
  phin=1-phip-phil;
}

public void phil(float theValue) {
  phil=theValue;
  phin=1-phip-phil;
}

public void Rad(float theValue) {
  R=theValue;
}

public void ClearShadow() {
  //clear Shadow array
  for (i=0; i<NP; i++) {
    for (int k=0; k<Res; k++) {
      Shadow[i][k]=0;
    }
  }
}

public void InitParticles() {
  //initialize particle positions
  for (i=0; i<NP; i++) {
    Particles[i][0] = XSize/2 + random(-1.0f, 1.0f)*200; //x-pos
    Particles[i][1] = YSize/2 + random(-1.0f, 1.0f)*200; //y-pos
    Particles[i][2] = vbase*cos(alphainit) + randomNormal()*Vel/10; //vx
    Particles[i][3] = vbase*sin(alphainit) + randomNormal()*Vel/10; //vy
  }
}

public void draw() {
  //set background image, load font, write title
    background(255);
    //frame.setTitle("Bird Flocking");
    ellipseMode(CENTER);
    noStroke();
  
  //centroid position
  if (DrawCentroid==true) {  
    tint(255, 50);
    image(imCH,xcentroid-24,ycentroid-24);
    noFill(); stroke(200); strokeWeight(3); ellipse(xcentroid,ycentroid,2*rcMax,2*rcMax);
  }
  
  //active particle coordinates
    xc=Particles[kAct][0]; yc=Particles[kAct][1]; 
    
  //DRAW SHADOWS and COMPUTE DISTANCES & CENTROID, CHECK MOUSE OVER
    noStroke(); xcentroid=0; ycentroid=0;
    for (i=0; i<NP; i++) {
      xp=Particles[i][0]; yp=Particles[i][1]; //particle coordinates
      xcentroid += xp; ycentroid += yp; 
      if (i!=kAct) {
        rp=sqrt((xp-xc)*(xp-xc)+(yp-yc)*(yp-yc)); //relative particle coordinates
        if (ShadowsToggle==true) {
          Theta=atan(xp-xc,yp-yc); alph=asin(R/rp); thetap=Theta+alph; thetam=Theta-alph; //relative angles
          fill(100);
          beginShape();
            vertex(xc+rp*cos(thetap),yc+rp*sin(thetap));
            vertex(xc+5000*cos(thetap),yc+5000*sin(thetap));
            vertex(xc+5000*cos(thetam),yc+5000*sin(thetam));
            vertex(xc+rp*cos(thetam),yc+rp*sin(thetam));
          endShape();
        }
        rpVec[i]=rp;
      } else {
        rpVec[i]=10000;
      }
      //needed for particle dragging
      if (RUN==false) {
        if (mouseX > xp-2*R && mouseX < xp+2*R && mouseY > yp-2*R && mouseY < yp+2*R) {
          bover = true;
          kover = i;  
        } else {
          boverFC += 1; 
        }
        if (boverFC == NP) {bover = false;} //mouse over no Particle
      }
    }
    xcentroid = xcentroid/NP; //get centroid x-coordinate
    ycentroid = ycentroid/NP; //get centroid y-coordinate
    
  //get index for sorted rp-array
    rpSort = sort(rpVec); //get sorted distance array
    for (i=0; i<NP; i++) {
      for (j=0; j<NP; j++) {
         if (rpVec[j] == rpSort[i]) {
           iorder[i] = j; //particle index of j-th-closest particle to kAct
         }
      }
    }
    
  //Generate discretized Shadow Array
    projanglex=0; projangley=0; opacity=0;
    rcMax=0; //maximum radius of the flock
    for (j=0; j<NP-1; j++) {
      //get index i of j-th particle from active particle
        i=iorder[j];
      //parameters
        xp=Particles[i][0]; yp=Particles[i][1]; //particle coordinates
        rp=sqrt((xp-xc)*(xp-xc)+(yp-yc)*(yp-yc)); //relative particle coordinates
        Theta=atan(xp-xc,yp-yc);
        if (rp<R) {
          alph=PI/2;
        } else {
          alph=asin(R/rp);
        }
      //shadow projection at each sorted distance index  
        if (FastToggle==false) { //get true shadow array
          thetap=Theta+alph; thetam=Theta-alph; //relative angles
          for (k=0; k<Res; k++) {
            thetak=k*dTheta; thetakp=thetak+dThetah; thetakm=thetak-dThetah;
            if ((((thetam > thetakm) && (thetam < thetakp)) || ((thetap > thetakm) && (thetap < thetakp))) || ((thetam < thetakm) && (thetap > thetakp))) {
              Shadow[j][k] = 1;
            } else {
              Shadow[j][k] = 0;
              if (j>0) {
                if (Shadow[j-1][k] == 1) {
                  Shadow[j][k] = 1; //i.e. accumulated shadow from a particle closer than the j-th particle
                }
              }
            }
          }
        } else { //quick and dirty shadow-mode / FAST mode
          projanglex += alph*cos(Theta);
          projangley += alph*sin(Theta);
          opacity += alph;
        }
      //find maximum radius of the flock
        rc = sqrt((xp-xcentroid)*(xp-xcentroid)+(yp-ycentroid)*(yp-ycentroid));
        if (rc>rcMax) {
           rcMax=rc;
        }
    }
    
  //(DRAW BIRDS), COMPUTE VISIBILITY AND NEAREST NEIGHBOR VELOCITY
    strokeWeight(1);
    vlocx = 0; vlocy = 0; iNNc=0; 
    for (j=0; j<NP; j++) {
      //get index of j-th particle, starting from the closest one
        i=iorder[j];
        iNN=false;
      //parameters of the j-th-closest particle with index i
        xp=Particles[i][0]; yp=Particles[i][1]; //particle coordinates
        veloangle = atan(Particles[i][2],Particles[i][3]); //particle velocity direction
      //check visibility only in slow-mode (accurate mode)
      if (FastToggle==false) {
        rp=sqrt((xp-xc)*(xp-xc)+(yp-yc)*(yp-yc)); //relative particle coordinates
        Theta=atan(xp-xc,yp-yc); alph=asin(R/rp); thetap=Theta+alph; thetam=Theta-alph; //relative angles
        //determine wether i!=kAct is visible from kAct
        if (i!=kAct) {
          //check visibility of Particle i from Particle kAct
          if (j >= 1) {
            kstart=max(0,PApplet.parseInt(thetam/dTheta));
            kend=min(PApplet.parseInt(thetap/dTheta),Res-1);
            SC=0;
            for (k=kstart; k<=kend; k++) {
              SC += Shadow[j-1][k];
            }
            if (SC == (kend-kstart+1)) {
              visible = false; //particle is invisible to kAct
            } else {
              visible = true; //particle is visible to kAct
              //check nearest neighbor
              if (iNNc < nNN) {
                iNN = true; //j-th particle (with index i) is a visible nearest neighbor
                iNNc += 1; //making sure at least nNN nearest neighbors are selected
              }
            }
          } else if (j==0) {
            visible = true; //closest particle must be visible to kAct
          }
        }
      } else {
        visible = true; // in fast mode all birds are visible
        if (j<nNN) {
          iNN=true; //simply consider nNN nearest neighbors 
        }
      }
      
      //is the particle a visible nearest neighbor (of nNN)? => local velocity
      if (iNN==true) {
          vlocx += Particles[i][2];
          vlocy += Particles[i][3];
      }
      
      if (FastToggle==false) {
        //DRAW BIRD'S VELOCITY vector as am arrow
        if (ArrowToggle==true) {
          stroke(0);
          arrow(xp,yp,xp+ArrLength*cos(veloangle),yp+ArrLength*sin(veloangle));
        }
      
        if (RUN==false | SpheresToggle==true) {
        //DRAW BIRD
          if (DrawBird==true) {
            //draw bird image!!
            if (i==kAct) {
              tint(255, 0, 0, 255); //active bird is redish
            } else if (visible==false) {
              tint(255, 70); //invisible bird is transparent
            } else if (iNN==true) {
              tint(0, 255, 0, 255); //nearest neighbor bird is green
            } else {
              tint(255, 255);
            }
            pushMatrix();
            translate(xp, yp);
            rotate(veloangle);
            image(imBIRD,-imBIRDWidthH,-imBIRDHeightH);
            popMatrix();
          } else if (SpheresToggle==true) {
           //drawing settings and local velocity based on visibility
            if (AnnotToggle==true) {
              //stroke options (boundary of the disc)
              if (visible==true) {
                if (iNN == true) {
                  stroke(0,255,0); //green boundary for visible nearest neighbor
                } else {
                  noStroke(); //no boundary for visible
                }
              } else {
                stroke(255,0,0); //no boundary for invisible
              }
              //filling options of the disc
              if (i==kAct) {
                fill(255,0,0); //active particle is a red filled circle
              } else {
                fill(0,0,(1-1.0f*j/NP)*255); //colorize particle blueish according to distance index
              }
            } else {
              noStroke(); fill(0); //no special annotation, all birds look equal and black
            } 
            ellipse(xp,yp,2*R,2*R); //draw simple spheres!!
          }
        }
      } else {
        noStroke(); fill(0); ellipse(xp,yp,2*R,2*R);
      }
      //fill(0); text(str(rpindex),xp+20,yp+4);
    }
  
  //GET AVERAGE NEAREST NEIGHBOR SPEED. GET RANDOM DIRECTION
      vlocx = vlocx/nNN; vlocy = vlocy/nNN; 
      vlocangle = atan(vlocx,vlocy);
      if (RUN==true) {
        randangle = random(0, TWO_PI);
      }
  
  //DRAW EQUALLY
  if (RUN==true && DrawBird==true) {
    for (i=0; i<NP; i++) {
      xp=Particles[i][0]; yp=Particles[i][1]; //particle coordinates
      veloangle = atan(Particles[i][2],Particles[i][3]);
      //draw bird image!!
        tint(255, 255);
        pushMatrix();
        translate(xp, yp);
        rotate(veloangle);
        image(imBIRD,-imBIRDWidthH,-imBIRDHeightH);
        popMatrix();
    }
  }  
    
  //DRAW PROJECTION CIRCLE and COMPUTE AVERAGE PROJECTION ANGLE
    stroke(0); strokeWeight(1); strokeJoin(ROUND); strokeCap(ROUND);
    if (FastToggle==false) { //draw true projection circle and average
      for (k=0; k<Res; k++) {
        thetak=k*dTheta;
        if (ProjToggle==true) {
          if (Shadow[NP-2][k]==1) {
            stroke(0); //black for shadow
          } else {
            stroke(230); //light gray if no shadow
          }
          line(xc+pR*cos(thetak-dThetah),yc+pR*sin(thetak-dThetah),xc+pR*cos(thetak+dThetah),yc+pR*sin(thetak+dThetah));
        }
        projanglex += Shadow[NP-2][k]*cos(thetak);
        projangley += Shadow[NP-2][k]*sin(thetak);
        opacity += Shadow[NP-2][k];
      }
    }
    projangle=atan(projanglex/opacity,projangley/opacity); //angle of average projection
    if (ProjToggle==true) {
      stroke(255,0,0); strokeWeight(3); //projection direction is RED
        line(xc+(pR-3)*cos(projangle),yc+(pR-3)*sin(projangle),xc+(pR+3)*cos(projangle),yc+(pR+3)*sin(projangle));
      stroke(0,255,0); strokeWeight(3); //nearest visible neaighbor direction is GREEN
        line(xc+(pR-3)*cos(vlocangle),yc+(pR-3)*sin(vlocangle),xc+(pR+3)*cos(vlocangle),yc+(pR+3)*sin(vlocangle));
      stroke(0,0,255); strokeWeight(3); //random direction is BLUE
        line(xc+(pR-3)*cos(randangle),yc+(pR-3)*sin(randangle),xc+(pR+3)*cos(randangle),yc+(pR+3)*sin(randangle));
    }
      
  //SIMULATE MOTION
    if (RUN==true) {
      //velocity update, projection velocity + local velocity + noise
        vkActx = phip*cos(projangle) + phin*cos(randangle) + phil*cos(vlocangle);
        vkActy = phip*sin(projangle) + phin*sin(randangle) + phil*sin(vlocangle);
        vkAct = sqrt(vkActx*vkActx + vkActy*vkActy);
        veloangle = atan(Particles[kAct][2], Particles[kAct][3]);
        Particles[kAct][2] = vbase*cos(veloangle) + Vel*(vkActx/vkAct); //vx
        Particles[kAct][3] = vbase*sin(veloangle) + Vel*(vkActy/vkAct); //vy
      //update position
        Particles[kAct][0] += Particles[kAct][2]; //x-pos
        Particles[kAct][1] += Particles[kAct][3]; //y-pos
      //reflecting boundary conditions
        xc = Particles[kAct][0]; yc = Particles[kAct][1];
        if (xc > XSize) {
          Particles[kAct][0] = xc - (xc-XSize);
        }
        if (xc < 0) {
          Particles[kAct][0] = -xc;
        }
        if (yc > YSize) {
          Particles[kAct][1] = yc - (yc-YSize);
        }
        if (yc < 0) {
          Particles[kAct][1] = -yc;
        }
      //re-center whole flock
        if ((xcentroid + rcMax) > XSize) {
          for (i=0; i<NP; i++) {
            Particles[i][0] -= (xcentroid-XSize/2);
          }
        }
        if ((xcentroid - rcMax) < 0) {
          for (i=0; i<NP; i++) {
            Particles[i][0] += XSize/2 - xcentroid;
          }
        }
        if ((ycentroid + rcMax) > YSize) {
          for (i=0; i<NP; i++) {
            Particles[i][1] -= (ycentroid-YSize/2);
          }
        }
        if ((ycentroid - rcMax) < 0) {
          for (i=0; i<NP; i++) {
            Particles[i][1] += YSize/2 - ycentroid;
          }
        }
    }     
    
  //advance active particle automatically by one in simulation mode
    fill(0);
    if (RUN==true) {
      kAct = (kAct + 1) % NP;
      if (kAct==0) {
         CylceC += 1;
      }
      text("time-step " + str(CylceC),XSize-80,11);
    }

  if (FastToggle==false) {    
  //formula images next to sliders
    tint(255, 255);
    image(imdv,10+bw+10,10);
    image(imphip,10+bw+10,30);
    image(imphil,10+bw+10,50);
    String phinStr = nf(phin, 1,4); //noise term
    text(phinStr,14,70+12);
    image(imphin,10+bw+10,70);
    image(imR,10+bw+10,93);
    image(imv0,10+bw+10,115);
  
  //Double click counter reduction by one (essentially a time variable)
    DoubleClick -= 1;
  
  //text  
    text("'SPACE': run/pause",10,YSize-87);
    text("'1': color discs",10,YSize-10);
    text("'2': shadows",10,YSize-21);
    text("'3': projection",10,YSize-32);
    text("'4': draw birds",10,YSize-43);
    text("'5': draw disc",10,YSize-54);
    text("'6': draw arrows",10,YSize-65);
    text("'7': draw centroid",10,YSize-76);
    text("'0': re-initialize",10,YSize-98);
    text("'f': fast mode",10,YSize-109);
  } else {
    fill(255,0,0);
    text("'f': fast mode",10,YSize-10);
  }
}

// return a random value in the range [-1..1] with normal distribution around 0.
// Implements the Marsaglia Polar Method, as described in wikipedia, but only returns one of the values.
public float randomNormal()
{
  float x = 1.0f, y = 1.0f, 
        s = 2.0f; // s = x^2 + y^2
  while(s >= 1.0f)
  {
    x = random(-1.0f, 1.0f);
    y = random(-1.0f, 1.0f);
    s = x*x + y*y;
  }
  return x * sqrt(-2.0f * log(s)/s);
}

public float atan(float tx, float ty) {
  float theta = abs(atan(ty/tx));
  
  if(tx < 0 && ty > 0) // Quadrant II
    theta = PI - theta;
  else if(tx < 0 && ty < 0) // Quadrant III
    theta = PI + theta;
  else if(tx > 0 && ty < 0) // Quadrant IV
    theta = TWO_PI - theta;
    
//  if(ty == 0 && tx > 0)
//    theta = 0;
//  else if(tx == 0 && ty > 0)
//    theta = PI / 2;
//  else if(tx == 0 && ty < 0)
//    theta = (3.0 / 2.0) * PI;
//  else if(ty == 0 && tx < 0)
//    theta = PI;
    
  return theta;
} 

public void arrow(float x1, float y1, float x2, float y2) {
  line(x1, y1, x2, y2);
  pushMatrix();
  translate(x2, y2);
  float a = atan2(x1-x2, y2-y1);
  rotate(a);
  line(0, 0, -2, -5);
  line(0, 0, 2, -5);
  popMatrix();
}

//change active Particle
public void keyPressed() {
  //switch actve particle
    if (key == RETURN || key == ENTER) {
      kAct = (kAct+1) % NP;
    }
  //SWITcH RUN / PAUSE  
    if (key == ' ') {
      RUN = ! RUN;
      if (RUN==true) {
        ShadowsToggle=false;
      }
    }
  //SWITcH Annotate  
    if (key == '1') {
      AnnotToggle = ! AnnotToggle;
    }
  //SWITcH Shadows  
    if (key == '2') {
      ShadowsToggle = ! ShadowsToggle;
    }
  //SWITcH Shadows  
    if (key == '3') {
      ProjToggle = ! ProjToggle;
    }  
  //SWITCH Draw Birds  
    if (key == '4') {
      DrawBird = ! DrawBird;
      if (DrawBird==true) {
        SpheresToggle=false;
      }
    }    
  //SWITCH Draw Spheres  
    if (key == '5') {
      SpheresToggle = ! SpheresToggle;
      if (SpheresToggle==true) {
        DrawBird=false;
      }
    }
  //SWITCH Draw Arrows  
    if (key == '6') {
      ArrowToggle = ! ArrowToggle;
    }  
  //SWITCH Draw Spheres  
    if (key == '7') {
      DrawCentroid = ! DrawCentroid;
    }
  //Re-initialize  
    if (key == '0') {
      //re-initial particle position
      ClearShadow();
      InitParticles();
    }
  //Fast-mode  
    if (key == 'f') {
      FastToggle = ! FastToggle;
    }
}

//Toggle Zoom-Image
public void mousePressed() {
  //Toggle Locked state
  if(bover) { 
    locked = true; 
  } else {
    locked = false;
  }
  bdifx = mouseX - Particles[kover][0]; 
  bdify = mouseY - Particles[kover][1];
  
  if (DoubleClick > 0) {
    for (i=0; i<NP; i++) {
      xp=Particles[i][0]; yp=Particles[i][1]; //particle coordinates
      if (mouseX > xp-2*R && mouseX < xp+2*R && mouseY > yp-2*R && mouseY < yp+2*R) {
        kAct = i;
      }
    }
  }
  DoubleClick = 10;
}

public void mouseDragged() {
  if(locked) {
    xp = mouseX-bdifx;
    yp = mouseY-bdify; 
    Particles[kover][0] = xp;
    Particles[kover][1] = yp;
  }
}

public void mouseReleased() {
  locked = false;
}

//void UpdateDensity() {
//  Density.loadPixels();
//  for (int x = 0; x < width; x++) {
//    for (int y = 0; y < height; y++) {
//      int loc = x + y * width;
//      for (i=0; i<NP; i++) {
//        if () {
//          locDensity += 1
//        }
//      }
//      //float r = red   (imBG.pixels[loc]);
//      //float g = green (imBG.pixels[loc]);
//      //float b = blue  (imBG.pixels[loc]);
//      Density.pixels[loc] = ;//color(r,g,b);
//    }
//  }
//  Density.updatePixels();
//}
  static public void main(String args[]) {
    PApplet.main(new String[] { "--bgcolor=#FFFFFF", "BirdFlocking" });
  }
}
